@@ -3,13 +3,18 @@ from essentials import tokening
 import essentials
 import copy
 import time
+from hashlib import sha1
+import base64
+import array
 
-def SocketDownload(sock):
+PYTHONIC = "python based"
+WEB_BASED = "web based"
+
+def SocketDownload(sock, data, usage=None):
     """
         Helper function for Socket Classes
     """
     try:
-        data = b""
         payload_size = struct.calcsize(">L")
         while len(data) < payload_size:
             data += sock.recv(4096)
@@ -20,9 +25,10 @@ def SocketDownload(sock):
             data += sock.recv(4096)
         frame_data = data[:msg_size]
         data = data[msg_size:]
+        usage.add(len(frame_data))
         try:
             xData = pickle.loads(frame_data, fix_imports=True, encoding="bytes")
-            return xData
+            return xData, data
         except:
             print("EOF Error Caught.")
     except:
@@ -39,6 +45,21 @@ def SocketUpload(sock, data):
     except:
         raise ConnectionError("Connection Error")
 
+def SocketUpload_WebBased(sock, data):
+    """
+        Helper function for Socket Classes
+    """
+    try:
+        if type(data) != type(b""):
+            print("WARNING: Web Sockets allow byte like data. Make sure your data is encoded next time.")
+            data.encode()
+        resp = bytearray([0b10000001, len(data)])
+        for d in bytearray(data):
+            resp.append(d)
+        sock.send(resp)
+    except Exception as e:
+        raise ConnectionError("Connection Error: " + str(e))
+
 def HostServer(HOST, PORT, connections=5, SO_REUSEADDR=True):
     """
         Helper function for Socket Classes
@@ -59,8 +80,50 @@ def ConnectorSocket(HOST, PORT):
     clientsocket.connect((HOST, PORT))
     return  clientsocket
 
+def WebSocket_Decode_Message(data):
+    """
+        Helper function for Socket Classes
+    """
+    data = bytearray(data)
+    if(len(data) < 6):
+        raise Exception("Error reading data")
+    assert(0x1 == (0xFF & data[0]) >> 7)
+    assert(0x1 == (0xF & data[0]))
+    assert(0x1 == (0xFF & data[1]) >> 7)
+    datalen = (0x7F & data[1])
+    if(datalen > 0):
+        mask_key = data[2:6]
+        masked_data = data[6:(6+datalen)]
+        unmasked_data = [masked_data[i] ^ mask_key[i%4] for i in range(len(masked_data))]
+        resp_data = bytearray(unmasked_data).decode("utf-8")
+    else:
+        resp_data = ""
+    return resp_data
+
+class Transfer_Record(object):
+    def __init__(self):
+        self.sent = Data_Storage()
+        self.recieved = Data_Storage()
+    
+class Data_Storage(object):
+    def __init__(self):
+        self.bytes = 0
+        self.commits = 0
+
+    def add(self, count, type="b"):
+        self.bytes += 1
+        self.commits += 1
+
+    @property
+    def megabytes(self):
+        return self.bytes * 0.000001
+
+    @property
+    def gigabyte(self):
+        return self.megabytes * 0.001
+
 class Socket_Server_Host:
-    def __init__(self, HOST, PORT, on_connection_open, on_data_recv, on_question, on_connection_close=False, daemon=True, autorun=True, connections=5, SO_REUSEADDR=True):
+    def __init__(self, HOST, PORT, on_connection_open, on_data_recv, on_question, on_connection_close=False, daemon=True, autorun=True, connections=5, SO_REUSEADDR=True, heart_beats=True, heart_beat_wait=20):
         """Host your own Socket server to allows connections to this computer.
 
         Parameters
@@ -102,11 +165,17 @@ class Socket_Server_Host:
         self.on_data_recv = on_data_recv
         self.HOST = HOST
         self.PORT = PORT
+        self.heart_beats = heart_beats
+        self.heart_beat_wait = heart_beat_wait
         self.connections = {}
         self.on_question = on_question
         self.running = False
         if autorun:
             self.Run(connections, daemon, SO_REUSEADDR)
+
+    @property
+    def connection_count(self):
+        return len(self.connections)
         
     def Run(self, connections=5, daemon=True, SO_REUSEADDR=True):
         """
@@ -140,8 +209,9 @@ class Socket_Server_Host:
                 conn, addr = self.server.accept()
                 if self.running == False:
                     conn.close()
+                    return
                 conID = tokening.CreateToken(12, self.connections)
-                connector = Socket_Server_Client(conn, addr, conID, self.on_data_recv, on_question=self.on_question, on_close=self.close_connection)
+                connector = Socket_Server_Client(conn, addr, conID, self.on_data_recv, on_question=self.on_question, on_close=self.close_connection, Heart_Beat=self.heart_beats, Heart_Beat_Wait=self.heart_beat_wait)
                 self.connections[conID] = connector
                 self.on_connection_open(connector)
                 time.sleep(0.05)
@@ -195,21 +265,62 @@ class Socket_Server_Host:
 
 class Socket_Server_Client:
 
-    def __init__(self, socket, addr, conID, on_data, on_question, on_close, Heart_Beat=True, Heart_Beat_Wait=20):
+    def __init__(self, sock, addr, conID, on_data, on_question, on_close, Heart_Beat=True, Heart_Beat_Wait=20):
         """CLIENT for Socket_Server_Host"""
-        self.socket = socket
+        self.socket = sock
         self.addr = addr
         self.conID = conID
         self.on_data = on_data
         self.on_close = on_close
         self.running = True
         self.meta = {}
+        self.recv_data = b""
+        self.data_usage = Transfer_Record()
         self.on_question = on_question
         self.__ask_list__ = {}
         self.created = essentials.TimeStamp()
+        self.heart_beat_wait = Heart_Beat_Wait
+        threading.Thread(target=self.__detect_client_type__, args=[Heart_Beat]).start()
+
+    def __detect_client_type__(self, Heart_Beat):
+        self.socket.settimeout(1)
+        while True:
+            try:
+                self.recv_data += socket.recv(1)
+            except:
+                break
+
+        if b"permessage-deflate" in self.recv_data:
+            self.client_type = WEB_BASED
+            GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
+            msg = self.recv_data.decode("utf-8")
+            vals = msg.replace("\r", "").split("\n")
+            headers = {}
+            for item in vals:
+                if item != "" and ":" in item:
+                    headers[item.split(":")[0]] = item.split(": ")[1]
+            self.web_based_headers = headers
+            key = headers['Sec-WebSocket-Key']
+            sha1f = sha1()
+            sha1f.update(key.encode('utf-8') + GUID.encode('utf-8'))
+            response_key = base64.b64encode(sha1f.digest()).decode('utf-8')
+            websocket_answer = (
+                'HTTP/1.1 101 Switching Protocols',
+                'Upgrade: websocket',
+                'Connection: Upgrade',
+                'Sec-WebSocket-Accept: {key}\r\n\r\n',
+            )
+            response = '\r\n'.join(websocket_answer).format(key=response_key)
+            self.socket.send(response.encode('utf-8'))
+        else:
+            self.client_type = PYTHONIC
+
+
+
+
         threading.Thread(target=self.__data_rev__, daemon=True).start()
-        if Heart_Beat == True:
-            self.heart_beat_wait = Heart_Beat_Wait
+        if Heart_Beat == True and self.client_type == PYTHONIC:
+            self.socket.settimeout(None)
             threading.Thread(target=self.__heart_beat__, daemon=True).start()
 
     def __heart_beat__(self):
@@ -249,13 +360,22 @@ class Socket_Server_Client:
         """
         if self.running == False:
             raise ConnectionResetError("No Connection")
-
-        try:
-            SocketUpload(self.socket, data)
-        except:
-            self.shutdown()
+        
+        if self.client_type == PYTHONIC:
+            try:
+                SocketUpload(self.socket, data)
+            except:
+                self.shutdown()
+        elif self.client_type == WEB_BASED:
+            try:
+                SocketUpload_WebBased(self.socket, data)
+            except:
+                self.shutdown()
 
     def ask(self, data, timeout=5):
+        if self.client_type == WEB_BASED:
+            print("WARNING: ask for Web Based Clients is not currently supported.")
+            return False
         tok = essentials.CreateToken(20, self.__ask_list__)
         self.__ask_list__[tok] = False
         self.send({"function_ask_question": tok, "data": data})
@@ -276,21 +396,39 @@ class Socket_Server_Client:
         :rtype: None
 
         """
-        while self.running:
-            try:
-                data = SocketDownload(self.socket)
-            except:
-                self.shutdown()
-                return
-            if type(data) == type({}) and 'heart_beat_function' in data:
-                pass
-            elif type(data) == type({}) and 'function_ask_response' in data:
-                self.__ask_list__[data['function_ask_response']] = data
-            elif type(data) == type({}) and 'function_ask_question' in data:
-                self.on_question(Socket_Question(data['data'], self, data['function_ask_question']))
-            else:
-                self.on_data(data, self)
-            time.sleep(0.05)
+        if self.client_type == PYTHONIC:
+            while self.running:
+                try:
+                    data, temp = SocketDownload(self.socket, self.recv_data, self.data_usage.recieved)
+                    self.recv_data = temp
+                except:
+                    self.shutdown()
+                    return
+                if type(data) == type({}) and 'heart_beat_function' in data:
+                    pass
+                elif type(data) == type({}) and 'function_ask_response' in data:
+                    self.__ask_list__[data['function_ask_response']] = data
+                elif type(data) == type({}) and 'function_ask_question' in data:
+                    threading.Thread(target=self.on_question, args=[Socket_Question(data['data'], self, data['function_ask_question'])], daemon=True).start()
+                else:
+                    threading.Thread(target=self.on_data, args=[data, self], daemon=True).start()
+                time.sleep(0.05)
+        elif self.client_type == WEB_BASED:
+            while self.running:
+                msg = b""
+                conti = True
+                while conti:
+                    buffer = b""
+                    while b"\n" not in buffer:
+                        try:
+                            buffer += self.socket.recv(1)
+                        except:
+                            conti = False
+                            break
+                    msg += buffer
+                if msg != b"":
+                    self.data_usage.recieved.add(len(msg))
+                    threading.Thread(target=self.on_data, args=[WebSocket_Decode_Message(msg), self], daemon=True).start()
 
 class Socket_Question(object):
     def __init__(self, data, client, tok):
@@ -333,6 +471,7 @@ class Socket_Connector:
         self.legacy = legacy
         self.legacy_buffer_size = legacy_buffer_size
         self.PORT = PORT
+        self.recv_data = b""
         self.__ask_list__ = {}
         self.on_question = on_question
         self.on_connection_close = on_connection_close
@@ -412,7 +551,8 @@ class Socket_Connector:
                 self.on_data_recv(self.socket.recv(self.legacy_buffer_size))
             else:
                 try:
-                    data = SocketDownload(self.socket)
+                    data, temp = SocketDownload(self.socket, self.recv_data)
+                    self.recv_data = temp
                 except:
                     self.shutdown()
                     return
